--[[ ToDO:
	
]]

--[[ Bugs
	
]]

-- Cooky 2017
-- Thanks to SoSo for his extensive help testing and UX expertise

------------------
-- Variable Set-up
------------------

local EBC = LibStub("AceAddon-3.0"):NewAddon("EndureBossChecker", "AceEvent-3.0", "AceComm-3.0", "AceTimer-3.0")
local LibSerializer = LibStub("AceSerializer-3.0")
local LibCompress = LibStub("LibCompress")
local LibEncodeTable = LibCompress:GetAddonEncodeTable()

local icon = LibStub("LibDBIcon-1.0", true)
local ldb = LibStub:GetLibrary("LibDataBroker-1.1")
local EBCLDB = ldb:NewDataObject("EndureBossChecker", {
	label = "EndureBossChecker",
	type = "data source",
	text = "EndureBossChecker",
	icon = "Interface\\AddOns\\EndureBossChecker\\Media\\icon",
})

-- Reset config every reload
local config = {
	console = false,
	version = "r1",
}

local PARENT_FRAME
local PARENT_FRAME_BASE_HEIGHT
local SCROLLING_OFFSET = 5
local BUTTON_HEIGHT = 30
local addon_prefix = "EBC_V"..config.version
local color_name = "|cffBF5FFFEndure|r Boss Checker"
local player_class_id = select(3,UnitClass("player"))

-- Defaults
local defaults = {
	version = config.version,
	minimapIcon = {hide = false, radius = 80, minimapPos = 45},
	selected_minimum_rank = 0,
	selected_raid_id = 768,
	selected_raid_name = "The Emerald Nightmare",
	info = {
	},
}

local ListOfRaidIDs = {768, 786, 875}

local ListOfBossIDs = {
	[768] = {1703, 1744, 1738, 1667, 1704, 1750, 1726},
	[786] = {1706, 1725, 1731, 1751, 1762, 1713, 1761, 1732, 1743, 1737},
	[875] = {2032, 2048, 2036, 2037, 2050, 2054, 2052, 2038, 2051},
}

local ListOfSlots = {"Head", "Neck", "Shoulder", "Back", "Chest", "Wrist", "Hands",
	"Waist", "Legs", "Feet", "Finger", "Trinket", "Relic"}

local RaidIDToName = {
	[768] = "The Emerald Nightmare",
	[786] = "The Nighthold",
	[875] = "Tomb of Sargeras",
}

-- Raid Id -> Boss Id
local BossIDToName = {
    [768] = { -- The Emerald Nightmare
        [1703] = "Nythendra",
        [1744] = "Elerethe Renferal",
        [1738] = "Il'gynoth, Heart of Corruption",
        [1667] = "Ursoc",
        [1704] = "Dragons of Nightmare",
        [1750] = "Cenarius",
        [1726] = "Xavius",
    },
    [786] = { -- The Nighthold
        [1706] = "Skorpyron",
        [1725] = "Chronomatic Anomaly",
        [1731] = "Trilliax",
        [1751] = "Spellblade Aluriel",
        [1762] = "Tichondrius",
        [1713] = "Krosus",
        [1761] = "High Botanist Tel'arn",
        [1732] = "Star Augur Etraeus",
        [1743] = "Grand Magistrix Elisande",
        [1737] = "Gul'dan",
    },
	[875] = { -- Tomb of Sargeras
		[2032] = "Goroth",
		[2048] = "Demonic Inquisition",
		[2036] = "Harjatan",
		[2037] = "Mistress Sassz'ine",
		[2050] = "Sisters of the Moon",
		[2054] = "The Desolate Host",
		[2052] = "Maiden of Vigilance",
		[2038] = "Fallen Avatar",
		[2051] = "Kil'jaeden",
	},
}

-- List of roles
local ListOfRoles = {
	"Dps",
	"Healer",
	"Tank"
}

local RoleIcons = {
	INLINE_DAMAGER_ICON,
	INLINE_HEALER_ICON,
	INLINE_TANK_ICON,
}

-- local RoleIcons = {
	-- 132152, -- Dps
	-- 136074, -- Healer
	-- 132110 -- Tank
-- }

local CacheNeed = {}

local ClassSpecIDList = {
	[1] = {71, 72, 73}, --Warrior
	[2] = {65, 66, 70}, --Paladin
	[3] = {253, 254, 255}, --Hunter
	[4] = {259, 260, 261}, --Rogue
	[5] = {256, 257, 258}, --Priest
	[6] = {250, 251, 252}, --DeathKnight
	[7] = {262, 263, 264}, --Shaman
	[8] = {62, 63, 64}, --Mage
	[9] = {265, 266, 267}, --Warlock
	[10] = {268, 269, 270}, --Monk
	[11] = {102, 103, 104, 105}, --Druid
	[12] = {577, 581}, --DemonHunter
}

--------------------
-- Utility Functions
--------------------

local round = function (num) return math.floor(num + .5) end

function EBC:RefreshMMButton()
	if EBC_DB.minimapIcon.hide then
		icon:Hide("EndureBossChecker")
	else
		icon:Show("EndureBossChecker")
	end
end

function EBCLDB:OnEnter()
	GameTooltip:SetOwner(self, "ANCHOR_NONE")
	GameTooltip:SetPoint("TOPLEFT", self, "BOTTOMLEFT")
	GameTooltip:ClearLines()
	
	GameTooltip:SetText(color_name, 1, 1, 1, true)
	GameTooltip:AddLine("|cffFF69B4Left-Click|r: Toggle window.", 1, 1, 1, 1)
	
	GameTooltip:Show()
end

function EBCLDB:OnClick(button)
	if button == "LeftButton" then
		if not PARENT_FRAME.firstOpened then
			if IsInGuild() then
				EBC:CreateGuildRankList()
				EBC:CreateRankButtons(PARENT_FRAME.guildSelectionFrame.minimumRankSelectionDropdown)
			else
				print(color_name..": Player must be in a guild to send and receive guild data.")
			end
			
			PARENT_FRAME.firstOpened = true
		end
			
		if PARENT_FRAME:IsVisible() then
			PARENT_FRAME:Hide()
		else
			PARENT_FRAME:Show()
		end
	end
end

function EBC:Console(message)
	if config.console then
		print(color_name..":", message)
	end
end

function EBC:CreateGuildRankList()
	self.list_guild_ranks = {}
	
	for i = 1, GetNumGuildMembers() do
		local _, rank_name, rank_id = GetGuildRosterInfo(i)
		self.list_guild_ranks[rank_id] = rank_name
	end
end

function EBC:ProcessChatCommand(command, action_args)
	if command == "" then
		if not PARENT_FRAME.firstOpened then
			if IsInGuild() then
				self:CreateGuildRankList()
				self:CreateRankButtons(PARENT_FRAME.guildSelectionFrame.minimumRankSelectionDropdown)
			else
				print(color_name..": Player must be in a guild to send and receive guild data.")
			end
			
			PARENT_FRAME.firstOpened = true
		end
		
		if PARENT_FRAME:IsVisible() then
			PARENT_FRAME:Hide()
		else
			PARENT_FRAME:Show()
		end
	elseif command == "help" then
		print(color_name, "options:")
		print("    /ebc reset - Reset your saved selections.")
		print("    /ebc version - Print addon version.")
		print("    /ebc minimap - Toggle showing the minimap button.")
	elseif command == "reset" then
		self:DBInit()
		self:Console("DB reset")
	elseif command == "version" then
		print(color_name.." version: "..EBC_DB.version)
	elseif command == "minimap" then
		EBC_DB.minimapIcon.hide = not EBC_DB.minimapIcon.hide
		EBC:RefreshMMButton()
	end
end

function EBC:CacheNeedInit()
	for _, raid_id in pairs(ListOfRaidIDs) do
		CacheNeed[raid_id] = {}
		
		for _, boss_id in pairs(ListOfBossIDs[raid_id]) do
			CacheNeed[raid_id][boss_id] = {}
			
			for role_id, _ in pairs(ListOfRoles) do
				CacheNeed[raid_id][boss_id][role_id] = {} -- {name, {table of specs}}
			end
		end
	end
end

function EBC:DBInit()
	EBC_DB = defaults
	EBC_DB.info.spec_id = GetSpecializationInfo(GetSpecialization())
	EBC_DB.info.role_id = 1
	
	if select(5,GetSpecializationInfoByID(EBC_DB.info.spec_id)) == "DAMAGER" then
		EBC_DB.info.role_id = 1
	elseif select(5,GetSpecializationInfoByID(EBC_DB.info.spec_id)) == "HEALER" then
		EBC_DB.info.role_id = 2
	elseif select(5,GetSpecializationInfoByID(EBC_DB.info.spec_id)) == "TANK" then
		EBC_DB.info.role_id = 3
	end

	for _, raid_id in pairs(ListOfRaidIDs) do
		EBC_DB.info[raid_id] = {}
		
		for _, boss_id in pairs(ListOfBossIDs[raid_id]) do
			EBC_DB.info[raid_id][boss_id] = {false, {}}
		end
	end
end

function EBC:ProcessVersion()
	if not EBC_DB.version or EBC_DB.version ~= config.version then
		self:DBInit()
		return
	end
end

------------------------
-- GUI Utility Functions
------------------------

function EBCHideAllDropdowns()
	local containers = {
		PARENT_FRAME.raidSelectionDropdown.container,
		PARENT_FRAME.personalSelectionFrame.specializationSelectionDropdown.container,
		PARENT_FRAME.personalSelectionFrame.roleSelectionDropdown.container,
		PARENT_FRAME.guildSelectionFrame.minimumRankSelectionDropdown.container,
	}
	
	for _, container in pairs(containers) do
		container:Hide()
	end
end

function EBCHideOtherDropdowns(self)
	if self.dropdownActive then
		self.container:Hide()
	else
		EBCHideAllDropdowns()
		self.container:Show()
	end
end

function EBC:Update(scrollFrame)
	local raid_id = scrollFrame.raid_id
	local scrollChildHeight = round(scrollFrame.scrollChild:GetHeight()) - SCROLLING_OFFSET
	local scrollFrameHeight = round(scrollFrame:GetHeight()+10) - SCROLLING_OFFSET
	local currPos = 0
	local offset = round(scrollFrame.scrollBar:GetValue()) - SCROLLING_OFFSET
	
	-- self:Console("Scrollframe: "..scrollFrameHeight.." Offset: "..offset.." Scrollchild: "..scrollChildHeight)
	
	for _, boss_id in pairs(ListOfBossIDs[raid_id]) do
		local button = self.scrollframebuttons[raid_id][boss_id]
		local buttonHeight = round(button:GetHeight()) + 5
		
		if currPos + buttonHeight <= offset then
			button:Hide()
			-- self:Console(BossIDToName[raid_id][boss_id].." hidden. #1 ("..buttonHeight..") ("..currPos + buttonHeight..")")
		elseif currPos + 0 < scrollFrameHeight + offset then
			button:Show()
			-- self:Console(BossIDToName[raid_id][boss_id].." shown. #2 ("..buttonHeight..") ("..currPos + buttonHeight..")")
		else
			button:Hide()
			-- self:Console(BossIDToName[raid_id][boss_id].." hidden. #3 ("..buttonHeight..") ("..currPos + buttonHeight..")")
		end
		
		currPos = currPos + buttonHeight
	end
	
	-- self:Console(" ")
end

function EBC:CheckedBoxForSpec(raid_id, boss_id)
	if not EBC_DB.info[raid_id][boss_id][1] then
		return false
	else
		local curr_spec_id = EBC_DB.info.spec_id
		
		for _, spec_id in pairs(EBC_DB.info[raid_id][boss_id][2]) do
			if spec_id == curr_spec_id then
				return true
			end
		end
		
		return false
	end
end

function EBC:SetDBBossCheck(checked, raid_id, boss_id, spec_id)
	if checked then
		if #EBC_DB.info[raid_id][boss_id][2] == 0 then
			EBC_DB.info[raid_id][boss_id][1] = true
			table.insert(EBC_DB.info[raid_id][boss_id][2], spec_id)
		else
			table.insert(EBC_DB.info[raid_id][boss_id][2], spec_id)
		end
	else
		if #EBC_DB.info[raid_id][boss_id][2] == 1 then
			EBC_DB.info[raid_id][boss_id][1] = false
			EBC_DB.info[raid_id][boss_id][2] = {}
		else
			for i = 1, #EBC_DB.info[raid_id][boss_id][2] do
				if EBC_DB.info[raid_id][boss_id][2][i] == spec_id then
					EBC_DB.info[raid_id][boss_id][2][i] = nil
					return
				end
			end
		end
	end
end

function EBC:SpecName(spec_id)
	return select(2, GetSpecializationInfoByID(spec_id))
end

function EBC:SpecIcon(spec_id)
	return select(4, GetSpecializationInfoByID(spec_id))
end

function EBC:SpecNameIcon(spec_id)
	local name = self:SpecName(spec_id)
	local icon = self:SpecIcon(spec_id)
	
	return "|T"..icon..":0|t "..name
end

---------------
-- Display Text
---------------

function EBC:CreateBossText()
	for _, raid_id in pairs (ListOfRaidIDs) do
		for _, boss_id in pairs (ListOfBossIDs[raid_id]) do
			local needCount = 0
			local temp = 5 + round(self.scrollframebuttons[raid_id][boss_id].text:GetHeight()) + 5
			local button = self.scrollframebuttons[raid_id][boss_id]
			button.clickedHeight = 5 + round(button.text:GetHeight()) + 5
			
			for role_id, role_name in pairs(ListOfRoles) do
				local fontString = _G[self.scrollframebuttons[raid_id][boss_id]:GetName().."Role"..role_id]
				local tempText = "|cffBF5FFF"..role_name.."|r: "
				
				local sortingTable = {}
				for name in pairs(CacheNeed[raid_id][boss_id][role_id]) do
					table.insert(sortingTable, name)
				end
				
				table.sort(sortingTable)
				
				for _, name in pairs(sortingTable) do
					-- self:Console(name)
					spec_table = CacheNeed[raid_id][boss_id][role_id][name]
					for i = 1, select(2, GetNumGuildMembers()) do
						local fullname = name.."-"..GetRealmName()
						
						if select(1, GetGuildRosterInfo(i)) == fullname then
							rank = select(3, GetGuildRosterInfo(i))
							break
						end
					end
					
					-- for i = 1, 10 do
					if rank <= EBC_DB.selected_minimum_rank then
						for _, spec_id in pairs(spec_table) do
							local iconId = select(4,GetSpecializationInfoByID(spec_id))
							tempText = tempText .. "|T"..iconId..":0|t"
						end
						
						tempText = tempText .. name .. " "
						needCount = needCount + 1
					end
					-- end
				end
				
				fontString:SetText(tempText)
				button.clickedHeight = button.clickedHeight + round(fontString:GetStringHeight()) + 5
			end
			
			self.scrollframebuttons[raid_id][boss_id].text:SetText(BossIDToName[raid_id][boss_id].." ("..needCount..")")
		end
	end
end

---------------------
-- Addon Communcation
---------------------

function EBC:EncodeAndSendData(target)
	local seralized_data = LibSerializer:Serialize(EBC_DB.info)
	local compressed_data = LibCompress:CompressHuffman(seralized_data)
	local encoded_data = LibEncodeTable:Encode(compressed_data)
	
	if target then
		self:SendCommMessage(addon_prefix, encoded_data, "WHISPER", target, "BULK")
	else
		self:SendCommMessage(addon_prefix, encoded_data, "GUILD", nil, "BULK")
	end
end

function EBC:OnCommReceived(prefix, data, _, sender)
	if prefix ~= addon_prefix then
		return
	end
	
	if data == "UPDATE" then
		self:EncodeAndSendData(sender)
	else
		local decoded_data = LibEncodeTable:Decode(data)
		
		local decompressed_data, error_message = LibCompress:Decompress(decoded_data)
		if not decompressed_data then
			self:Console("Error decompressing: " .. error_message)
			return
		end
		
		local success, final_data = LibSerializer:Deserialize(decompressed_data)
		if not success then
			self:Console("Error deseralizing: " .. final_data)
			return
		end
		
		self:ProcessData(final_data, sender)
		self:CreateBossText()
	end
end

function EBC:ProcessData(data, sender)
	self:RemoveNameFromCache(sender)
	
	for _, raid_id in pairs(ListOfRaidIDs) do
		for _, boss_id in pairs(ListOfBossIDs[raid_id]) do
			local need_table = data[raid_id][boss_id]
			local role_id = data.role_id
			
			local split_specs = {
				{nil, {}}, -- Dps
				{nil, {}}, -- Healer
				{nil, {}}, -- Tanks
			}
			
			if need_table[1] then
				for _, spec_id in pairs(need_table[2]) do
					if select(5,GetSpecializationInfoByID(spec_id)) == "DAMAGER" then
						split_specs[1][1] = sender
						table.insert(split_specs[1][2], spec_id)
					elseif select(5,GetSpecializationInfoByID(spec_id)) == "HEALER" then
						split_specs[2][1] = sender
						table.insert(split_specs[2][2], spec_id)
					elseif select(5,GetSpecializationInfoByID(spec_id)) == "TANK" then
						split_specs[3][1] = sender
						table.insert(split_specs[3][2], spec_id)
					end
				end
			end
			
			if split_specs[role_id][1] then
				self:InputNameAndSpecs(role_id, raid_id, boss_id, split_specs[role_id])
			end
		end
	end
end

function EBC:RemoveNameFromCache(name)
	-- modify for new format of specs included?
	for _, raid_id in pairs(ListOfRaidIDs) do
		for _, boss_id in pairs(ListOfBossIDs[raid_id]) do
			for role_id, _ in pairs(ListOfRoles) do
				CacheNeed[raid_id][boss_id][role_id][name] = nil
			end
		end
	end
end

function EBC:InputNameAndSpecs(role_id, raid_id, boss_id, spec_table)
	local name = spec_table[1]
	CacheNeed[raid_id][boss_id][role_id][name] = {}
	CacheNeed[raid_id][boss_id][role_id][name] = spec_table[2]
	table.sort(CacheNeed[raid_id][boss_id][role_id][name])
end

function EBC:RequestUpdate(source)
	if source then
		self:SendCommMessage(addon_prefix, "UPDATE", "WHISPER", source)
	else
		self:SendCommMessage(addon_prefix, "UPDATE", "GUILD")
	end
end

function EBC:LoggedOff(source)
	self:RemoveNameFromCache(source)
	self:CreateBossText()
end

---------------
-- GUI Creation
---------------

function EBC:SelectRaidSelectionContainer(container_list, target)
	for id, container in pairs(container_list) do
		container:Hide()
	end
	container_list[target]:Show()
	
	PARENT_FRAME:SetHeight(PARENT_FRAME_BASE_HEIGHT + self.checkbox_container_height[target])
end

function EBC:SelectScrollframe(scrollframe_list, target)
	for id, scrollframe in pairs(scrollframe_list) do
		scrollframe:Hide()
	end
	
	scrollframe_list[target]:Show()
end

function EBC:SetCheckboxesBySpec()
	for _, raid_id in pairs(ListOfRaidIDs) do
		for _, boss_id in pairs(ListOfBossIDs[raid_id]) do
			self.checkboxes[raid_id][boss_id]:SetChecked(self:CheckedBoxForSpec(raid_id, boss_id))
		end
	end
end

function EBC:CreateRaidButtons(dropdown)
	dropdown.text:SetText(RaidIDToName[EBC_DB.selected_raid_id])
	local prev_button = nil
	local height = 0
	for i, raid_id in pairs(ListOfRaidIDs) do
		local b = CreateFrame("Button", dropdown.container:GetName().."Button"..i, dropdown.container, "EBC_DropdownObjectTemplate")
		
		b.text:SetText(i..": "..RaidIDToName[raid_id])
		b:SetWidth(dropdown.container:GetWidth()-4)
		
		b:HookScript("OnClick", function()
			EBC_DB.selected_raid_id = raid_id
			EBC_DB.selected_raid_name = RaidIDToName[raid_id]
			
			dropdown.text:SetText(RaidIDToName[raid_id])
			dropdown.container:Hide()
			
			self:SelectRaidSelectionContainer(self.checkbox_containers, raid_id)
			self:SelectScrollframe(self.scrollframes, raid_id)
		end)
		
		if prev_button then
			b:SetPoint("TOPLEFT", prev_button, "BOTTOMLEFT", 0, 0)
		else
			b:SetPoint("TOPLEFT", dropdown.container, "TOPLEFT", 2, -2)
		end
		
		prev_button = b
		height = height + round(b:GetHeight())
	end
	dropdown.container:SetHeight(height+4)
end

function EBC:CreateSpecializationButtons(dropdown)
	dropdown.text:SetText(self:SpecNameIcon(EBC_DB.info.spec_id))
	local prev_button = nil
	local height = 0
	for i, spec_id in pairs(ClassSpecIDList[player_class_id]) do
		local b = CreateFrame("Button", dropdown.container:GetName().."Button"..i, dropdown.container, "EBC_DropdownObjectTemplate")
		
		b.text:SetText(self:SpecNameIcon(spec_id))
		b:SetWidth(dropdown.container:GetWidth()-4)
		
		b:HookScript("OnClick", function()
			EBC_DB.info.spec_id = spec_id
			
			dropdown.text:SetText(b.text:GetText())
			dropdown.container:Hide()
			self:SetCheckboxesBySpec()
		end)
		
		if prev_button then
			b:SetPoint("TOPLEFT", prev_button, "BOTTOMLEFT", 0, 0)
		else
			b:SetPoint("TOPLEFT", dropdown.container, "TOPLEFT", 2, -2)
		end
		
		prev_button = b
		height = height + round(b:GetHeight())
	end
	dropdown.container:SetHeight(height+4)
end

function EBC:CreateRoleButtons(dropdown)
	dropdown.text:SetText(RoleIcons[EBC_DB.info.role_id].." "..ListOfRoles[EBC_DB.info.role_id])
	-- dropdown.text:SetText(RoleIcons[EBC_DB.info.role_id]..ListOfRoles[EBC_DB.info.role_id])
	local prev_button = nil
	local height = 0
	for i = 1,3 do
		local b = CreateFrame("Button", dropdown.container:GetName().."Button"..i, dropdown.container, "EBC_DropdownObjectTemplate")
		
		b.text:SetText(RoleIcons[i].." "..ListOfRoles[i])
		-- b.text:SetText(RoleIcons[i]..ListOfRoles[i])
		b:SetWidth(dropdown.container:GetWidth()-4)
		
		b:HookScript("OnClick", function()
			EBC_DB.info.role_id = i
			
			dropdown.text:SetText(b.text:GetText())
			dropdown.container:Hide()
		end)
		
		if prev_button then
			b:SetPoint("TOPLEFT", prev_button, "BOTTOMLEFT", 0, 0)
		else
			b:SetPoint("TOPLEFT", dropdown.container, "TOPLEFT", 2, -2)
		end
		
		prev_button = b
		height = height + round(b:GetHeight())
	end
	
	dropdown.container:SetHeight(height+4)
end

function EBC:CreateRankButtons(dropdown)
	dropdown.text:SetText(self.list_guild_ranks[EBC_DB.selected_minimum_rank])
	local prev_button = nil
	local height = 0
	for i = 0, #self.list_guild_ranks do
		local b = CreateFrame("Button", dropdown.container:GetName().."Button"..i, dropdown.container, "EBC_DropdownObjectTemplate")
		
		b.text:SetText((i+1)..": "..self.list_guild_ranks[i])
		b:SetWidth(dropdown.container:GetWidth()-4)
		
		b:HookScript("OnClick", function()
			EBC_DB.selected_minimum_rank = i
			
			dropdown.text:SetText(self.list_guild_ranks[i])
			dropdown.container:Hide()
			
			self:CreateBossText()
		end)
		
		if prev_button then
			b:SetPoint("TOPLEFT", prev_button, "BOTTOMLEFT", 0, 0)
		else
			b:SetPoint("TOPLEFT", dropdown.container, "TOPLEFT", 2, -2)
		end
		
		prev_button = b
		height = height + round(b:GetHeight())
	end
	
	dropdown.container:SetHeight(height+4)
end

function EBC:CreatePersonalSelection()
	self.checkbox_containers = {}
	self.checkbox_container_height = {}
	self.checkboxes = {}
	for i, raid_id in pairs(ListOfRaidIDs) do
		local raid_container = CreateFrame("Frame", PARENT_FRAME.personalSelectionFrame.personalSelectionContainer:GetName().."CheckboxContainer"..raid_id, 
			PARENT_FRAME.personalSelectionFrame.personalSelectionContainer, "EBC_DisplayContainerTemplate")
		raid_container:SetPoint("TOPLEFT", raid_container:GetParent(), "TOPLEFT")
		raid_container:SetPoint("BOTTOMRIGHT", raid_container:GetParent(), "BOTTOMRIGHT")
		raid_container.label:SetText(RaidIDToName[raid_id])
		
		self.checkbox_containers[raid_id] = raid_container
		self.checkbox_container_height[raid_id] = 5
		self.checkboxes[raid_id] = {}
		
		local prev_checkbox = nil
		for j, boss_id in pairs(ListOfBossIDs[raid_id]) do
			local checkbox = CreateFrame("CheckButton", raid_container:GetName().."Checkbox"..boss_id, raid_container, "EBC_CheckButtonTemplate")
			checkbox.text:SetText(BossIDToName[raid_id][boss_id])
			checkbox:SetChecked(self:CheckedBoxForSpec(raid_id, boss_id))
			
			self.checkboxes[raid_id][boss_id] = checkbox
			
			checkbox:HookScript("OnClick", function()
				self:SetDBBossCheck(checkbox:GetChecked(), raid_id, boss_id, EBC_DB.info.spec_id)
			end)
			
			if prev_checkbox then
				checkbox:SetPoint("TOPLEFT", prev_checkbox, "BOTTOMLEFT", 0, -5)
			else
				checkbox:SetPoint("TOPLEFT", raid_container, "TOPLEFT", 5, -5)
			end
			prev_checkbox = checkbox
			
			self.checkbox_container_height[raid_id] = self.checkbox_container_height[raid_id] + round(checkbox:GetHeight()) + 5
		end
	end
end

function EBC:CreateRaidSelection()
	self.scrollframes = {}
	self.scrollframebuttons = {}
	for _, raid_id in pairs(ListOfRaidIDs) do
		local scrollframe = CreateFrame("ScrollFrame", PARENT_FRAME.guildSelectionFrame.guildSelectionContainer:GetName().."ScrollFrame"..raid_id,
			PARENT_FRAME.guildSelectionFrame.guildSelectionContainer, "EBC_ScrollFrameTemplate")
		scrollframe:SetPoint("TOPLEFT", scrollframe:GetParent(), "TOPLEFT", 0, -SCROLLING_OFFSET)
		scrollframe:SetPoint("BOTTOMRIGHT", scrollframe:GetParent(), "BOTTOMRIGHT", 0, SCROLLING_OFFSET)
		scrollframe.label:SetText(RaidIDToName[raid_id])
		
		scrollframe.scrollBar:SetMinMaxValues(SCROLLING_OFFSET, SCROLLING_OFFSET)
		scrollframe.scrollBar:SetValue(SCROLLING_OFFSET)
		self.scrollframes[raid_id] = scrollframe
		self.scrollframebuttons[raid_id] = {}
		
		local child = CreateFrame("Frame", scrollframe:GetName().."ScrollChild", scrollframe)
		scrollframe.scrollChild = child
		scrollframe.raid_id = raid_id
		child:SetSize(260, (#ListOfBossIDs[raid_id]*30)+SCROLLING_OFFSET)
		
		scrollframe.scrollBar:HookScript("OnValueChanged", function(self)
			EBC:Update(scrollframe)
		end)
		
		prevButton = nil
		for k, boss_id in pairs(ListOfBossIDs[raid_id]) do
			local button = CreateFrame("Button", child:GetName().."Button"..boss_id, child, "EBC_ScrollChildOptionTemplate")
			button.text:SetText(BossIDToName[raid_id][boss_id])
			
			self.scrollframebuttons[raid_id][boss_id] = button
			
			button.unclickedHeight = round(button:GetHeight())
			button.clickedHeight = 0
			
			button:HookScript("OnClick", function(self)
				local diff = 0
				local currmin, currmax = scrollframe.scrollBar:GetMinMaxValues()
				
				if not self.active then
					for _, boss_id in pairs(ListOfBossIDs[raid_id]) do
						local button = EBC.scrollframebuttons[raid_id][boss_id]
						
						if (IsShiftKeyDown() or button == self) and not button.active then
							diff = diff + (button.clickedHeight - button.unclickedHeight)
							
							button:SetBackdropColor(1, 1, 1, 0.15)
							button.text:SetPoint("TOPLEFT", button, "TOPLEFT", 5, -5)
							button.arrow:SetTexture("Interface\\AddOns\\EndureBossChecker\\Media\\arrowup.blp")
							button.arrow:SetPoint("TOPRIGHT", button, "TOPRIGHT", -5, -5)
							
							button:SetHeight(button.clickedHeight)
							button.role1:Show()
							button.role2:Show()
							button.role3:Show()
							
							button.active = true
						end
					end
					
					scrollframe.scrollBar:SetMinMaxValues(round(currmin), round(currmax + diff))
					child:SetHeight(round(child:GetHeight()) + diff)
				else
					for _, boss_id in pairs(ListOfBossIDs[raid_id]) do
						local button = EBC.scrollframebuttons[raid_id][boss_id]
						
						if (IsShiftKeyDown() or button == self) and button.active then
							diff = diff + (button.clickedHeight - button.unclickedHeight)
							
							button:SetBackdropColor(1, 1, 1, 0.025)
							button.text:SetPoint("LEFT", button, "LEFT", 5, 0)
							button.arrow:SetTexture("Interface\\AddOns\\EndureBossChecker\\Media\\arrowdown.blp")
							button.arrow:SetPoint("TOPRIGHT", button, "TOPRIGHT", -5, -10)
							
							button:SetHeight(button.unclickedHeight)
							button.role1:Hide()
							button.role2:Hide()
							button.role3:Hide()
							
							button.active = false
						end
					end
					
					scrollframe.scrollBar:SetMinMaxValues(round(currmin), round(currmax - diff))
					child:SetHeight(round(child:GetHeight()) - diff)
				end
				
				EBC:Update(scrollframe)
			end)
			
			if not prevButton then
				button:SetPoint("TOPLEFT", child, "TOPLEFT", 5, -5)
			else
				button:SetPoint("TOPLEFT", prevButton, "BOTTOMLEFT", 0, -5)
			end
			
			prevButton = button
		end
		scrollframe:SetScrollChild(scrollframe.scrollChild)
	end
end

function EBC:PARENT_FRAMEInit()
	-- Create All Dropbox Options
	self:CreateRaidButtons(PARENT_FRAME.raidSelectionDropdown)
	self:CreateSpecializationButtons(PARENT_FRAME.personalSelectionFrame.specializationSelectionDropdown)
	self:CreateRoleButtons(PARENT_FRAME.personalSelectionFrame.roleSelectionDropdown)
	
	-- Main Display Containers
	self:CreatePersonalSelection()
	self:SelectRaidSelectionContainer(self.checkbox_containers, EBC_DB.selected_raid_id)
	
	self:CreateRaidSelection()
	self:SelectScrollframe(self.scrollframes, EBC_DB.selected_raid_id)
	
	-- Script Finalization
	PARENT_FRAME.personalSelectionFrame:HookScript("OnHide", function()
		if IsInGuild() then
			self:EncodeAndSendData()
		else
			print(color_name..": Player must be in a guild to send and receive guild data.")
		end
	end)
	
	PARENT_FRAME.guildSelectionFrame:HookScript("OnShow", function()
		self:CreateBossText()
	end)
	
	PARENT_FRAME.personalSelectionButton:HookScript("OnClick", function()
		PARENT_FRAME.personalSelectionButton:SetBackdropColor(0, 0.55, 0.85, 0.6)
		PARENT_FRAME.guildSelectionButton:SetBackdropColor(0, 0.55, 0.85, 0.3)
	end)
	
	PARENT_FRAME.guildSelectionButton:HookScript("OnClick", function()
		if IsInGuild() then
			PARENT_FRAME.personalSelectionFrame:Hide()
			PARENT_FRAME.guildSelectionFrame:Show()
			EBCHideAllDropdowns()
			PARENT_FRAME.guildSelectionButton:SetBackdropColor(0, 0.55, 0.85, 0.6)
			PARENT_FRAME.personalSelectionButton:SetBackdropColor(0, 0.55, 0.85, 0.3)
		else
			print(color_name..": Player must be in a guild to send and receive guild data.")
		end
	end)
	
	PARENT_FRAME.personalSelectionButton:HookScript("OnEnter", function()
		if not PARENT_FRAME.personalSelectionFrame:IsVisible() then
			PARENT_FRAME.personalSelectionButton:SetBackdropColor(0, 0.55, 0.85, 1)
		end
	end)
	
	PARENT_FRAME.guildSelectionButton:HookScript("OnEnter", function()
		if not PARENT_FRAME.guildSelectionFrame:IsVisible() then
			PARENT_FRAME.guildSelectionButton:SetBackdropColor(0, 0.55, 0.85, 1)
		end
	end)
	
	PARENT_FRAME.personalSelectionButton:HookScript("OnLeave", function()
		if PARENT_FRAME.personalSelectionFrame:IsVisible() then
			PARENT_FRAME.personalSelectionButton:SetBackdropColor(0, 0.55, 0.85, 0.6)
		else
			PARENT_FRAME.personalSelectionButton:SetBackdropColor(0, 0.55, 0.85, 0.3)
		end
	end)
	
	PARENT_FRAME.guildSelectionButton:HookScript("OnLeave", function()
		if PARENT_FRAME.guildSelectionFrame:IsVisible() then
			PARENT_FRAME.guildSelectionButton:SetBackdropColor(0, 0.55, 0.85, 0.6)
		else
			PARENT_FRAME.guildSelectionButton:SetBackdropColor(0, 0.55, 0.85, 0.3)
		end
	end)
end

------------------------------------------
-- Addon Initilization and Event functions
------------------------------------------

function EBC_PARENT_FRAMELoaded()
	PARENT_FRAME = EBCParentFrame
	PARENT_FRAME_BASE_HEIGHT = PARENT_FRAME:GetHeight()
	
	EBC:CacheNeedInit()
	
	EBC:RegisterEvent("PLAYER_LOGIN")
	
	-- Make slash commands
	SLASH_EBC1 = "/ebc"
	SlashCmdList["EBC"] = function(origmsg)
		origmsg = string.lower(origmsg)
		local action_args = {}
		
		for action in origmsg:gmatch("%w+") do
			table.insert(action_args, action)
		end
		
		local command = action_args[1] or origmsg
		
		EBC:ProcessChatCommand(command, action_args)
	end
	
	-- Set up recurring table garbage collection (the use of this many 
	-- tables ends up causing a lot of garbage tables to accumulate lolololol)
	local function cleanup()
		EBC:ScheduleTimer(function() 
			collectgarbage(collect) 
			cleanup()
		end, 60)
	end
	
	-- Finished loading
	print(color_name,"loaded, type /ebc for options.")
end

function EBC:PLAYER_LOGIN()
	if not EBC_DB then
		self:DBInit()
	end
	self:ProcessVersion()
	
	self:RegisterComm(addon_prefix, "OnCommReceived")
	self:RegisterEvent("CHAT_MSG_SYSTEM")
	self:RegisterEvent("GUILD_RANKS_UPDATE")
	
	self:RequestUpdate()
	self:PARENT_FRAMEInit()
	
	if not icon:IsRegistered("EndureBossChecker") then
		icon:Register("EndureBossChecker", EBCLDB, EBC_DB.minimapIcon)
	end
end

function EBC:CHAT_MSG_SYSTEM(_,message)
	-- Guild Logging On
	if message and string.find(message, "has come online.") then
		local startIndex = string.find(message, "%[") + 1
		local endIndex = string.find(message, "%]") - 1
		source = string.sub(message, startIndex, endIndex)
		
		self:ScheduleTimer("RequestUpdate", 5, source)
	end
	
	-- Guild Logging Off
	if message and string.find(message, "has gone offline.") then
		local startIndex = 1
		local endIndex = string.find(message, "has gone offline") - 2
		source = string.sub(message, startIndex, endIndex)
		
		self:LoggedOff(source)
	end
end

function EBC:GUILD_RANKS_UPDATE()
	self:CreateBossText()
end